(require 'asdf)
(require 'pal)

(defpackage :boulderdash
  (:use :common-lisp)
  (:export #:play))

(in-package :boulderdash)

(defparameter *dirt-image* nil)
(defparameter *wall-image* nil)
(defparameter *rock-image* nil)
(defparameter *diamond-image* nil)
(defparameter *door-image* nil)
(defparameter *player-image* nil)
(defparameter *victory-image* nil)
(defparameter *death-image* nil)

(defun load-images ()
  (setf *dirt-image* (pal:load-image "Images/dirt.png")
        *wall-image* (pal:load-image "Images/wall.png")
        *rock-image* (pal:load-image "Images/rock.png")
        *diamond-image* (pal:load-image "Images/diamond.png")
        *door-image* (pal:load-image "Images/door.png")
        *player-image* (pal:load-image "Images/player.png")
        *victory-image* (pal:load-image "Images/victory.png")
        *death-image* (pal:load-image "Images/death.png")))

(defclass level ()
  ((width :initarg :width :accessor width)
   (height :initarg :height :accessor height)
   (data :initarg :data :accessor data)
   (entry-position :initarg :entry-position :accessor entry-position)
   (exit-position :initarg :exit-position :accessor exit-position)
   (diamond-count :initarg :diamond-count :accessor diamond-count)))

(defun char->element (char)
  (ecase char
    (#\space 'empty)
    (#\# 'wall)
    (#\- 'dirt)
    (#\o 'rock)
    (#\* 'diamond)
    (#\@ 'entry)
    (#\& 'exit)))

(defun parse-map (lines)
  (let* ((width (length (first lines)))
         (height (length lines))
         (data (make-array (list height width)))
         (entry nil)
         (exit nil)
         (diamond-count 0))
    (loop for y below height
          for line in lines
          do (loop for x below width
                   as element = (char->element (aref line x))
                   do (setf (aref data y x) element)
                   when (eq element 'entry)
                     do (setf entry (cons x y))
                        (setf (aref data y x) 'empty)
                   when (eq element 'exit)
                     do (setf exit (cons x y))
                   when (eq element 'diamond)
                     do (incf diamond-count)))
    (make-instance 'level
                   :width width
                   :height height
                   :data data
                   :entry-position entry
                   :exit-position exit
                   :diamond-count diamond-count)))

(defun load-map (level)
  (let ((filename (format nil "Levels/level~2,'0d.txt" level)))
    (with-open-file (stream filename)
      (loop as line = (read-line stream nil nil)
            until (null line)
            collect line into lines
            finally (return (parse-map lines))))))

(defun handle-falling-objects (level falling-objects player-position)
  (loop for y from (- (height level) 2) downto 0
        do (loop for x below (width level)
                 when (member (aref (data level) y x) '(rock diamond))
                 do (cond ((and (eq (aref (data level) (1+ y) x) 'empty)
                                (or (/= x (car player-position))
                                    (/= (1+ y) (cdr player-position))
                                    (aref falling-objects y x)))
                           (rotatef (aref (data level) y x)
                                    (aref (data level) (1+ y) x))
                           (setf (aref falling-objects y x) nil
                                 (aref falling-objects (1+ y) x) t))
                          ((and (eq (aref (data level) y (1- x)) 'empty)
                                (eq (aref (data level) (1+ y) (1- x)) 'empty)
                                (member (aref (data level) (1+ y) x)
                                        '(rock diamond))
                                (or (/= (1- x) (car player-position))
                                    (and (/= (1+ y) (cdr player-position))
                                         (/= y (cdr player-position)))))
                           (rotatef (aref (data level) y x)
                                    (aref (data level) y (1- x)))
                           (setf (aref falling-objects y x) nil
                                 (aref falling-objects y (1- x)) t))
                          ((and (eq (aref (data level) y (1+ x)) 'empty)
                                (eq (aref (data level) (1+ y) (1+ x)) 'empty)
                                (member (aref (data level) (1+ y) x)
                                        '(rock diamond))
                                (or (/= (1+ x) (car player-position))
                                    (and (/= (1+ y) (cdr player-position))
                                         (/= y (cdr player-position)))))
                           (rotatef (aref (data level) y x)
                                    (aref (data level) y (1+ x)))
                           (setf (aref falling-objects y x) nil
                                 (aref falling-objects y (1+ x)) t))
                          (t
                           (setf (aref falling-objects y x) nil))))))

(defparameter +tile-size+ 32)

(defparameter +screen-tile-width+ 16)
(defparameter +screen-tile-height+ 15)

(defparameter +screen-width+ (* +screen-tile-width+ +tile-size+))
(defparameter +screen-height+ (* +screen-tile-height+ +tile-size+))

(defun draw-game (level player-position status)
  (pal:clear-screen 0 0 0)
  (loop for y from (- (cdr player-position) (floor +screen-tile-height+ 2))
        for screen-y from 0
        repeat +screen-tile-height+
        do (loop for x from (- (car player-position) (floor +screen-tile-width+ 2))
                 for screen-x from 0
                 repeat +screen-tile-width+
                 when (and (< -1 x (width level))
                           (< -1 y (height level)))
                 do (let ((element (aref (data level) y x)))
                      (unless (eq element 'empty)
                        (pal:draw-image (case element
                                          (wall *wall-image*)
                                          (dirt *dirt-image*)
                                          (rock *rock-image*)
                                          (diamond *diamond-image*)
                                          (exit *door-image*))
                                        (pal:v (* screen-x +tile-size+)
                                               (* screen-y +tile-size+)))))
                 do (when (and (= x (car player-position))
                               (= y (cdr player-position)))
                      (pal:draw-image (ecase status
                                        (:playing *player-image*)
                                        (:victory *victory-image*)
                                        (:dead *death-image*))
                                      (pal:v (* screen-x +tile-size+)
                                             (* screen-y +tile-size+)))))))

(defun play (level)
  (pal:with-pal (:width +screen-width+ :height +screen-height+
                 :fullscreenp nil :fps 50)
    (load-images)
    (let* ((level (load-map level))
           (player-position (copy-tree (entry-position level)))
           (diamond-count 0)
           (falling-objects (make-array (list (height level) (width level))
                                        :initial-element nil))
           (status :playing)
           (frame 0))
      (pal:event-loop ()
        (when (= (mod frame 5) 0)
          (let ((next-position (copy-tree player-position)))
            (pal:test-keys
              (:key-up (decf (cdr next-position)))
              (:key-down (incf (cdr next-position)))
              (:key-left (decf (car next-position)))
              (:key-right (incf (car next-position)))
              (:key-q (return-from play)))
            (let ((element (aref (data level)
                                 (cdr next-position) (car next-position))))
              (when (eq status :playing)
                (when (member element '(empty dirt diamond exit))
                  (setf player-position next-position)
                  (when (eq element 'diamond)
                    (incf diamond-count))
                  (unless (eq element 'exit)
                    (setf (aref (data level)
                                (cdr next-position) (car next-position))
                          'empty)))))))
        (when (= (mod frame 5) 2)
          (handle-falling-objects level falling-objects player-position))
        (when (eq (aref (data level)
                        (cdr player-position) (car player-position))
                  'rock)
          (setf status :dead))
        (when (and (equal player-position (exit-position level))
                   (>= diamond-count (diamond-count level)))
          (setf status :victory))
        (draw-game level player-position status)
        (incf frame)))))
